package com.ae7q.communication;

import java.io.*;
import java.net.*;
import java.util.logging.*;
import java.util.prefs.*;

import com.ultimeth.util.*;

//==================================================================================================
/** AgwpeServlet is an AX.25 packet message servlet for AGWPE.
 *
 *  @version    2006.04.09
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
public class AgwpeServlet extends RadioServlet {
    private static final int          MAX_XMIT    = 128;
    private static final byte[]       MESSAGE_ACK = "<?ack?>".getBytes();  // { 0x06 }
    public static  final Preferences  preferences = Preferences.systemNodeForPackage( RadioServlet.class )
                                                              .node( "agwpe" );
    public static  final AgwpeServlet servlet     = new AgwpeServlet();
           private       AgwpeSocket  agwpeSocket = null;
           private       int          agwpePort;
           private       String       inetHost;
           private       int          inetPort;
           private       String       userName;
           private       String       password;

    //==============================================================================================
    public enum Preference {
        StationId     ( "" ),
        MessagePath   ( "" ),
        AgwpeTcpIpHost( "localhost" ),
        AgwpeTcpIpPort( "8000" ),
        AgwpeRadioPort( "0" ),
        AgwpeLogin    ( "" ),
        AgwpePassword ( "" ),
        ;
        public final String defaultValue;

        //------------------------------------------------------------------------------------------
        /** Preference constructor.
         *
         *  @param  defaultValue        the default value for the configuration parameter.
         */
        Preference( String defaultValue ){
            this.defaultValue = defaultValue;
        } // AgwpeServlet.Preference constructor

        //------------------------------------------------------------------------------------------
        /** Gets a preference value.
         *
         *  @return String              - the value.
         */
        public String get(){
            return preferences.get( name(), defaultValue );
        } // AgwpeServlet.Preference.get

        //------------------------------------------------------------------------------------------
        /** Sets a preference value.
         *
         *  @param  value               the value to save.
         */
        public void put( String value ){
            if ( value != null ){
                preferences.put( name(), value );
            }
        } // AgwpeServlet.Preference.put
    } // AgwpeServlet.Preference enum


    //----------------------------------------------------------------------------------------------
    /** Construct a AgwpeServlet object.
     */
    AgwpeServlet() {
        open();
    } // AgwpeServlet constructor

    //----------------------------------------------------------------------------------------------
    /** Checks the validity of the configuration.
     *
     *  @return boolean                 - the state of the configuration.
     */
    public boolean isValid(){
        try {
            localCall   =              Preference.StationId.get().toUpperCase();
            messagePath =              Preference.MessagePath.get();
            inetHost    =              Preference.AgwpeTcpIpHost.get();
            inetPort    = new Integer( Preference.AgwpeTcpIpPort.get() );
            agwpePort   = new Integer( Preference.AgwpeRadioPort.get() );
            userName    =              Preference.AgwpeLogin.get();
            password    =              Preference.AgwpePassword.get();
            if ( localCall.equals( "" ) ){
                throw new Exception( "Missing 'StationId' configuration value" );
            }
            if ( ! new File( messagePath ).isDirectory() ){
                throw new Exception( "Invalid 'MessagePath' configuration value" );
            }
            for ( State state : State.values() ){
                File file = new File( messagePath, state.toString() );
                if ( ! file.isDirectory()  &&  ! file.mkdir() ){
                    throw new Exception( "Unable to create '" + state + "' subdirectory" );
                }
             }
            return true;
        } catch ( Exception exception ){
            log.log( Level.CONFIG, "Invalid configuration", exception );
        }
        return false;
    } // AgwpeServlet.isValid

    //----------------------------------------------------------------------------------------------
    /** Open the AGWPE connection.
     *
     *  @return boolean                 - the state of the AGWPE connection.
     */
    public boolean open(){
        try {
            if ( isValid() ){
                if ( agwpeSocket == null ){
                    rxConnect.clear();
                    agwpeSocket = new AgwpeSocket( inetHost, inetPort );
                    connect();
                }
                return true;
            }
        } catch ( Exception exception ){
            log.log( Level.SEVERE, "Error starting servlet instance", exception );
        }
        return false;
    } // AgwpeServlet.open

    //----------------------------------------------------------------------------------------------
    /** Connect to an AGWPE connection.
     *
     *  @throws UnknownHostException    on an unknown host error.
     *  @throws IOException             on an I/O error.
     */
    private void connect()                              throws  UnknownHostException,
                                                                IOException {
        AgwpeSocket.Segment segment;

        agwpeSocket.send( new AgwpeSocket.Segment( agwpePort, AgwpeSocket.Segment.Type.VERSION, 0,
                                                   null, null, null ) );
        segment = agwpeSocket.receive();
        log.info( "AGWPE version: " + AgwpeSocket.Segment.extract (segment.body, 0 ) + "."
                                    + AgwpeSocket.Segment.extract (segment.body, 4 ) );
        if ( ! userName.equals( "" ) ){
            byte[] body = new byte[ 510 ];
            AgwpeSocket.Segment.insert( userName, body, 0 );
            AgwpeSocket.Segment.insert( password, body, 255 );
            agwpeSocket.send( new AgwpeSocket.Segment( agwpePort, AgwpeSocket.Segment.Type.LOGIN,  0,
                                                       null, null, body ) );
        }
        agwpeSocket.send( new AgwpeSocket.Segment( agwpePort, AgwpeSocket.Segment.Type.UNREGISTER, 0,
                                                   localCall, null, null ) );
        agwpeSocket.send( new AgwpeSocket.Segment( agwpePort, AgwpeSocket.Segment.Type.REGISTER,   0,
                                                   localCall, null, null ) );
        segment = agwpeSocket.receive();
        if ( segment.body[ 0 ] != 0 ){
            log.info( localCall + " registration succeeded" );
            new Thread(){
                public void run(){
                    listen();
                }
            }.start();
            return;
        }
        throw new IOException( localCall + " registration failed" );
    } // AgwpeServlet connect

    //----------------------------------------------------------------------------------------------
    /** Get the ID of the connection.
     *
     *  @return String                  - the state of the connection, or null if not open.
     */
    public String getId(){
        return agwpeSocket != null ? inetHost + ":" + inetPort : null;
    } // AgwpeServlet.getId

    //----------------------------------------------------------------------------------------------
    /** Close the connection.
     */
    public synchronized void close(){
        if ( agwpeSocket != null ){
            try {
                agwpeSocket.close();
                agwpeSocket = null;
            } catch ( Exception exception ){
                log.log( Level.WARNING, "Error stopping servlet instance", exception );
            }
        }
    } // AgwpeServlet.close
    //----------------------------------------------------------------------------------------------
    /** Listen for AGWPE message segments.
     */
    void listen(){
        AgwpeSocket.Segment segment;
        PipedOutputStream   oStream;

        try {
            agwpeSocket.send( new AgwpeSocket.Segment( agwpePort, AgwpeSocket.Segment.Type.PORTS_INFO,     0,
                                                       null, null, null ) );
            agwpeSocket.send( new AgwpeSocket.Segment( agwpePort, AgwpeSocket.Segment.Type.MONITOR_TOGGLE, 0,
                                                       null, null, null ) );
            agwpeSocket.send( new AgwpeSocket.Segment( agwpePort, AgwpeSocket.Segment.Type.HEARD,          0,
                                                       null, null, null ) );
            while ( agwpeSocket != null ){
                segment = agwpeSocket.receive();
                switch ( segment.type ){
                case AgwpeSocket.Segment.Type.PORTS_INFO:
                    log.config( "[" + segment.type + "] " + segment.body() );
                    continue;
                case AgwpeSocket.Segment.Type.HEARD:
                    if ( segment.from.trim().length() == 0 ){
                        continue;
                    }
                    break;
                case AgwpeSocket.Segment.Type.CONNECT:
                    if ( segment.body().startsWith( "*** CONNECTED To Station " ) ){
                        final PipedInputStream iStream    = new PipedInputStream();
                        final String           remoteCall = segment.from;
                        rxConnect.put( remoteCall, new PipedOutputStream( iStream ) );
                        new Thread(){
                            public void run(){
                                receiveFile( remoteCall, iStream );
                            }
                        }.start();
                    } else if ( segment.body().startsWith( "*** CONNECTED With " ) ){
                        final PipedInputStream iStream    = new PipedInputStream();
                        final String           remoteCall = segment.from;
                        rxConnect.put( remoteCall, new PipedOutputStream( iStream ) );
                        new Thread(){
                            public void run(){
                                sendFile( remoteCall, iStream );
                            }
                        }.start();
                    } else {
                        agwpeSocket.send( new AgwpeSocket.Segment( agwpePort, AgwpeSocket.Segment.Type.DISCONNECT, 0,
                                                                   localCall, segment.from, null ) );
                        log.warning( "Unknown connection type: " + segment.body() );
                    }
                    log.fine( "[" + segment.type + "] " + segment.body() );
                    continue;
                case AgwpeSocket.Segment.Type.DISCONNECT:
                    oStream = rxConnect.remove( segment.from );
                    if ( oStream != null ){
                        oStream.flush();
                        oStream.close();
                    }
                    log.fine( "[" + segment.type + "] " + segment.body() );
                    //  ! "*** DISCONNECTED RETRYOUT With "
                    if ( segment.body().startsWith( "*** DISCONNECTED From Station " )
                          &&  getMatchingFileName( segment.from ) != null ){
                        startSend( segment.from );
                    }
                    continue;
                case AgwpeSocket.Segment.Type.CONNECTED_DATA:
                    oStream = rxConnect.get( segment.from );
                    if ( oStream != null ){
                        oStream.write( segment.body );
                        continue;
                    }
                    break;
                case AgwpeSocket.Segment.Type.MONITORED_FRAME:
                case AgwpeSocket.Segment.Type.INFO_TRAME:
                case AgwpeSocket.Segment.Type.SUPERVISORY_FRAME:
                case AgwpeSocket.Segment.Type.RAW_FRAME:
                case AgwpeSocket.Segment.Type.UNCONNECTED_FRAME:
                    break;
                }
                log.fine( "[" + segment.type + "] " + segment.body() );
            }
        } catch ( Exception exception ){
            log.log( Level.SEVERE, "Communications error", exception );
        }
        close();
    } // AgwpeServlet.listen

    //----------------------------------------------------------------------------------------------
    /** Get an appropriately-sized output buffer.
     *
     *  @return byte[]                          - the byte array to use.
     */
    byte[] buffer(){
        return new byte[ MAX_XMIT ];
    } // AgwpeServlet.buffer

    //----------------------------------------------------------------------------------------------
    /** Start a connection to a remote station.
     *
     *  @param  remoteCall              the remote station callsign and SSID.
     *  @throws IOException             on an I/O error.
     */
    public void startSend( String remoteCall )          throws  IOException {
        if ( rxConnect.get( remoteCall ) == null ){
            agwpeSocket.send( new AgwpeSocket.Segment( agwpePort, AgwpeSocket.Segment.Type.CONNECT, 0,
                                                       localCall, remoteCall, null ) );
        }
    } // AgwpeServlet.startSend

    //------------------------------------------------------------------------------------------
    /** Send an outgoing message segment.
     *
     *  @param  remoteCall                      the callsign used for the search.
     *  @param  body                            the message segment body.
     *  @throws Exception                       on an error.
     */
    void sendSegment( String remoteCall, byte[] body )  throws  Exception {
        agwpeSocket.send( new AgwpeSocket.Segment( agwpePort, AgwpeSocket.Segment.Type.CONNECTED_DATA, 0,
                                                   localCall, remoteCall, body ) );
    } // AgwpeServlet.sendSegment

    //----------------------------------------------------------------------------------------------
    /** Terminates a sent message segment.
     *
     *  @param  remoteCall                      the callsign used for the search.
     *  @param  rxStream                        the receive InputStream for the connection.
     *  @throws Exception                       on an error.
     */
    void endSend( String remoteCall, InputStream rxStream )
                                                        throws  Exception {
        log.fine( "ETX sent, waiting for ACK" );
        if ( new ByteInputBuffer( rxStream ).waitFor( null, MESSAGE_ACK ) == null ){
            throw new EOFException( "From remote source: " + remoteCall );
        }
        log.fine( "ACK received" );
    } // AgwpeServlet.endSend

    //----------------------------------------------------------------------------------------------
    /** Terminates a received message.
     *
     *  @param  remoteCall                      the callsign used for the search.
     *  @throws Exception                       on an error.
     */
    void endReceive( String remoteCall )                throws  Exception {
        sendSegment( remoteCall, MESSAGE_ACK );
        log.fine( "ACK sent" );
    } // AgwpeServlet.endReceive
} // class AgwpeServlet
