package com.ae7q.communication;

import java.io.*;
import java.net.*;
import java.util.*;

import com.ultimeth.util.*;

//==================================================================================================
/** AgwpeSocket is an AX.25 packet class for AGWPE.
 *
 *  @version    2006.04.09
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
public class AgwpeSocket extends Socket {
    private OutputStream outputStream = getOutputStream();
    private InputStream  inputStream  = getInputStream();

    //----------------------------------------------------------------------------------------------
    /** Create the underlying socket.
     *
     *  @param  inetHost                the hostname or IP address for AGWPE.
     *  @param  inetPort                the port number for AGWPE.
     *  @throws UnknownHostException    on an unknown host error.
     *  @throws IOException             on an I/O error.
     */
    public AgwpeSocket( String inetHost, int inetPort )         throws  UnknownHostException,
                                                                        IOException {
        super( InetAddress.getByName( inetHost ), inetPort );
    } // AgwpeSocket constructor

    //----------------------------------------------------------------------------------------------
    /** Close the underlying socket.  Of dubious value.
     */
    public void finalize()                                      throws  Throwable {
        close();
    } // AgwpeSocket.finalize

    //----------------------------------------------------------------------------------------------
    /** Read the input stream until the buffer is full.
     *
     *  @param  buffer                  the buffer to be filled.
     *  @throws IOException             on an I/O error.
     */
    private void read( byte[] buffer )                          throws  IOException {
        for ( int index = 0;
                   index < buffer.length;
                   index += inputStream.read( buffer, index, buffer.length - index ) );
    } // AgwpeSocket.read

    //----------------------------------------------------------------------------------------------
    /** Receive an AGWPE message segment.
     *
     *  @return Segment                 - the received data.
     *  @throws IOException             on an I/O error.
     */
    public Segment receive()                                    throws  IOException {
        byte[] agwpeHeader = new byte[ Segment.HEADER_SIZE ];
        read( agwpeHeader );
        Segment segment = new Segment( agwpeHeader );
        if ( segment.body != null ){
            read( segment.body );
        }
        return segment;
    } // AgwpeSocket.receive

    //----------------------------------------------------------------------------------------------
    /** Send an AGWPE message segment.
     *
     *  @param  segment                 the message segment to be sent.
     *  @throws IOException             on an I/O error.
     */
    public void send( Segment segment )                         throws  IOException {
        outputStream.write( segment.header() );
        if ( segment.body != null ){
            outputStream.write( segment.body );
        }
    } // AgwpeSocket.send

    //==============================================================================================
    /** Represents an AGWPE message segment.
     */
    public static class Segment {
        public static final int HEADER_SIZE = 36;
        public static final int CALLSIGN_SIZE = 10;
        public static final int MAX_RELAYS = 7;
        int    port;
        char   type;
        int    pid;
        String from;
        String to;
        byte[] body;

        //------------------------------------------------------------------------------------------
        /** Extract a little-endian 16-bit integer value from a buffer.
         *
         *  @param  buffer              the buffer.
         *  @param  offset              the offset of the integer value.
         *  @return int                 - the integer value.
         */
        static int extract( byte[] buffer, int offset ){
            return (buffer[ offset ] & 0xFF) + ((buffer[ offset + 1 ] & 0xFF) << 8);
        } // AgwpeSocket.Segment.extract

        //------------------------------------------------------------------------------------------
        /** Insert a little-endian 16-bit integer value into a buffer.
         *
         *  @param  value               the integer value.
         *  @param  buffer              the buffer.
         *  @param  offset              the offset for the integer value.
         */
        static void insert( int value, byte[] buffer, int offset ){
            buffer[ offset + 0 ] = (byte)  value;
            buffer[ offset + 1 ] = (byte) (value >> 8);
        } // AgwpeSocket.Segment.insert

        //------------------------------------------------------------------------------------------
        /** Insert a String value into a buffer.
         *
         *  @param  string              the string.
         *  @param  buffer              the buffer.
         *  @param  offset              the offset for the string.
         */
        static void insert( String string, byte[] buffer, int offset ){
            // string.getBytes( 0, string.length(), buffer, offset );  -- deprecated
            for ( int index = 0;  index < string.length();  index++ ){
                buffer[ index + offset ] = (byte) string.charAt( index );
            }
        } // AgwpeSocket.Segment.insert

        //------------------------------------------------------------------------------------------
        /** Strip from a string any data after a trailing null byte.
         *
         *  @param  string              the string to be stripped.
         *  @return String              - the stripped string.
         */
        static String strip( String string ){
            int index = string.indexOf( 0 );
            return index < 0 ? string : string.substring( 0, index );
        } // AgwpeSocket.Segment.strip

        //------------------------------------------------------------------------------------------
        /** Build a byte array of AGWPE relay stations.
         *
         *  @param  string              a string of relay station names.
         *  @return byte[]              - the relay stations.
         */
        static byte[] buildRelays( String string ){
            StringTokenizer tokens  = new StringTokenizer( string.toUpperCase(), ", " );
            int             count   = tokens.countTokens();
            byte[]          relays  = null;
    
            if ( count > 0 ){
                relays = new byte[ 1 + count * AgwpeSocket.Segment.CALLSIGN_SIZE ];
                relays[ 0 ] = (byte) count;
                for ( int index = 1;  index < relays.length;
                                      index += AgwpeSocket.Segment.CALLSIGN_SIZE ){
                    AgwpeSocket.Segment.insert( tokens.nextToken(), relays, index );
                }
            }
            return relays;
        } // AgwpeSocket.Segment.buildRelays

        //------------------------------------------------------------------------------------------
        /** Construct an outgoing message segment.
         *
         *  @param  port                the AGWPE port.
         *  @param  type                the AGWPE segment type.
         *  @param  pid                 the AX.25 PID.
         *  @param  from                the sending station.
         *  @param  to                  the receiving station.
         *  @param  body                the segment body.
         */
        public Segment( int port, char type, int pid, String from, String to, byte[] body ){
            this.port = port;
            this.type = type;
            this.pid  = pid;
            this.from = from;
            this.to   = to;
            this.body = body;
        } // AgwpeSocket.Segment constructor

        //------------------------------------------------------------------------------------------
        /** Construct an incoming message segment.
         *
         *  @param  header              the raw header of the segment.
         */
        public Segment( byte[] header ){
            int length =    extract( header, 28);
            port =          extract( header, 0);
            type = (char)   extract( header, 4);
            pid  =          extract( header, 6);
            from = strip( new String( header,  8, CALLSIGN_SIZE ) );
            to   = strip( new String( header, 18, CALLSIGN_SIZE ) );
            body = new byte[ length ];
        } // AgwpeSocket.Segment constructor
            //------------------------------------------------------------------------------------------
        /** Build an outgoing raw header.
         *
         *  @return  byte[]             - the byte array containing the header.
         */
        public byte[] header(){
            byte[] header = new byte[ HEADER_SIZE ];
    
            insert( port, header, 0 );
            insert( type, header, 4 );
            insert( pid,  header, 6 );
            if ( from != null ){
                insert( from, header, 8 );
            }
            if ( to != null ){
                insert( to, header, 18 );
            }
            if ( body != null ){
                insert( body.length, header, 28 );
            }
            return header;
        } // AgwpeSocket.Segment.header

        //------------------------------------------------------------------------------------------
        /** Get the text portion of a message segment.
         *
         *  @return  String             -- the text portion of the segment.
         */
        String body(){
            return strip( new String( body ) );
        } // AgwpeSocket.Segment.body

        //==========================================================================================
        /** Message segment types.
         */
        interface Type {
            public static final char LOGIN               = 'P';
            public static final char VERSION             = 'R';
            public static final char PORTS_INFO          = 'G';
            public static final char PORT_CAPS           = 'g';
            public static final char REGISTER            = 'X';
            public static final char UNREGISTER          = 'x';
            public static final char PORT_FRAMES         = 'y';
            public static final char CONN_FRAMES         = 'Y';
            public static final char HEARD               = 'H';
            public static final char MONITOR_TOGGLE      = 'm';
            public static final char MONITOR_RAW_TOGGLE  = 'k';
            public static final char TRANSMIT            = 'M';
            public static final char TRANSMIT_VIA        = 'V';
            public static final char CONNECT             = 'C';
            public static final char CONNECT_VIA         = 'v';
            public static final char CONNECT_NONSTD      = 'c';
            public static final char CONNECTED_DATA      = 'D';
            public static final char DISCONNECT          = 'd';
            public static final char UNCONNECTED_FRAME   = 'U';
            public static final char INFO_TRAME          = 'I';
            public static final char SUPERVISORY_FRAME   = 'S';
            public static final char MONITORED_FRAME     = 'T';
            public static final char RAW_FRAME           = 'K';
        } // interface AgwpeSocket.Segment.Type
    } // class AgwpeSocket.Segment
} // class AgwpeSocket
