package com.ultimeth.util;

import java.io.*;
//import java.util.*;        // Name conflict w/ java.util.logging.*
import java.util.logging.*;  // Name conflict w/ java.util.*
import java.text.*;

//==================================================================================================
/** An abstract logging capability.
 *
 *  @version    2005.08.25
 *  @author     Copyright © 2002-2005 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
public class Log {
    public static final DateFormat isoDateFormat = new SimpleDateFormat( "yyyy-MM-dd HH:mm:ss" );
    public static final String     endOfLine     = System.getProperty( "line.separator" );

    //----------------------------------------------------------------------------------------------
    /** Initializes a Logger with a default date format.
     *
     *  @param  classObject             a class object to derive a Logger name from, or null.
     *  @return Logger                  - the requested Logger.
     */
    public static Logger get( Class classObject ){
        Logger logger = classObject != null
                      ? Logger.getLogger( classObject.getCanonicalName() )
                      : Logger.getAnonymousLogger();
        logger.setLevel( Level.ALL );
        return logger;
    }// Log.get

    //==============================================================================================
    /** An abstract Logger Handler.
     */
    abstract public static class Handler extends java.util.logging.Handler {
        private Formatter formatter;
        private Level     level = Level.ALL;

        //------------------------------------------------------------------------------------------
        /** Initializes a Logger.
         */
        public Handler(){
            setFormatter( null );
        }// Log.Handler constructor

        //------------------------------------------------------------------------------------------
        /** Tests if the handler would log at the indicated level.
         *
         *  @param  level               the level to test.
         *  @return Level               - whether the handler would log.
         */
        public boolean isLoggable( Level level ){
            return level.intValue() >= getLevel().intValue();
        }// Log.Handler.getLevel

        //------------------------------------------------------------------------------------------
        /** Gets the Handler's Level (Handler override of stupid, STUPID Sun security code).
         *
         *  @return Level               - the Handler's Level.
         */
        public Level getLevel(){
            return level;
        }// Log.Handler.getLevel

        //------------------------------------------------------------------------------------------
        /** Sets the Handler's Level (Handler override of stupid, STUPID Sun security code).
         *
         *  @param  level               the Level to use.
         */
        public void setLevel( Level level ){
            this.level = level;
            if ( getFormatter() instanceof Format ){
                ((Format) getFormatter()).setLogLevel( level );
            }
        }// Log.Handler.setLevel

        //------------------------------------------------------------------------------------------
        /** Gets the Handler's Formatter (Handler override of stupid, STUPID Sun security code).
         *
         *  @return Formatter           - the Handler's Formatter (never null).
         */
        public Formatter getFormatter(){
            return formatter;
        }// Log.Handler.getFormatter

        //------------------------------------------------------------------------------------------
        /** Sets the Handler's Formatter (Handler override of stupid, STUPID Sun security code).
         *
         *  @param  formatter           the Formatter to use.
         */
        public void setFormatter( Formatter formatter ){
            this.formatter = formatter != null ? formatter : new Format( isoDateFormat );
        }// Log.Handler.setFormatter

        //------------------------------------------------------------------------------------------
        /** Flushes the Handler (Handler override).
         */
        public void flush(){
        }// Log.Handler.flush

        //------------------------------------------------------------------------------------------
        /** Closes the Handler (Handler override).
         */
        public void close(){
        }// Log.Handler.close
    }// Log.Handler class

    //==============================================================================================
    /** A simple Logger Formatter.
     */
    public static class Format extends Formatter {
        public  final DateFormat dateFormat;
        private       Level      logLevel    = Level.ALL,
                                 detailLevel = Level.FINE;

        //------------------------------------------------------------------------------------------
        /** Initializes a Formatter with a date format.
         *
         *  @param  dateFormat          the log file DateFormat.
         */
        public Format( DateFormat dateFormat ){
            this.dateFormat  = dateFormat;
        }// Log.Format constructor

        //------------------------------------------------------------------------------------------
        /** Formats a LogRecord for logging (Formatter override).
         *
         *  @param  logRecord           the LogRecord to format.
         *  @return String              - the formatted LogRecord.
         */
        public String format( LogRecord logRecord ){
            StringBuffer logData = new StringBuffer( 1000 );
            Throwable    thrown  = logRecord.getThrown();

            logData.append( dateFormat.format( new java.util.Date( logRecord.getMillis() ) ) );
            logData.append( " [" + logRecord.getLevel() + "] " );
            logData.append( logRecord.getMessage() + endOfLine );
            if ( thrown != null ){
                logData.append( thrown + endOfLine );
                if ( getLogLevel().intValue() <= getDetailLevel().intValue() ){
                    for ( StackTraceElement stackTraceElement : thrown.getStackTrace() ){
                        logData.append( "    at" + stackTraceElement + endOfLine );
                    }
                }
            }
            return logData.toString();
        }// Log.Format.format

        //------------------------------------------------------------------------------------------
        /** Gets the Formatter's logging Level.
         *
         *  @return Level               - the Formatter's logging Level.
         */
        public Level getLogLevel(){
            return logLevel;
        }// Log.Handler.getLogLevel

        //------------------------------------------------------------------------------------------
        /** Sets the Formatter's logging Level.
         *
         *  @param  logLevel            the logging Level to use.
         */
        public void setLogLevel( Level logLevel ){
            this.logLevel = logLevel;
        }// Log.Format.setLogLevel

        //------------------------------------------------------------------------------------------
        /** Gets the Formatter's detail Level.
         *
         *  @return Level               - the Formatter's detail Level.
         */
        public Level getDetailLevel(){
            return detailLevel;
        }// Log.Handler.getDetailLevel

        //------------------------------------------------------------------------------------------
        /** Sets the Formatter's detail Level.
         *
         *  @param  detailLevel         the detail Level to use.
         */
        public void setDetailLevel( Level detailLevel ){
            this.detailLevel = detailLevel;
        }// Log.Format.setDetailLevel
    }// Log.Format class
}// Log class
