package com.ultimeth.util;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.util.jar.*;
import java.util.logging.*;
import java.util.prefs.*;
import javax.swing.*;

//==================================================================================================
/** Tools
 *
 *  @version    2005.08.25
 *  @author     Copyright © 2002 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
public class Tools {
    public static final URL     classUrl         = Tools.class.getResource( "Tools.class" );
    public static final boolean isJarFile        = classUrl.getProtocol().equals( "jar" );
    public static final String  JAR_PREFIX       = "jar:!/";
    public static final String  filePrefix       = isJarFile ? JAR_PREFIX : "./" ;
    public static final UIManager.LookAndFeelInfo[]
                                lookAndFeelInfos = UIManager.getInstalledLookAndFeels();

    //----------------------------------------------------------------------------------------------
    /** Returns the string with the first letter upper case, and the remainder lower case.
     *
     *  @param  name                    the String to use.
     *  @return String                  - the initial-cased String.
     */
    public static String initialCase( String name ){
        return name.substring( 0, 1 ).toUpperCase()
             + name.substring( 1    ).toLowerCase();
    }// Tools.initialCase

    //----------------------------------------------------------------------------------------------
    /** Updates a path name string for subsequent use.
     *
     *  @param  pathName                the pathname to format.
     *  @return String                  - the formatted pathname.
     */
    public static String formatPath( String pathName ){
        if ( pathName.length() > 2  &&  pathName.charAt( 0 ) == '/'
                                    &&  pathName.charAt( 2 ) == ':' ){
            pathName = pathName.substring( 1 );
        }
        if ( ! pathName.endsWith( "/" )  &&  ! pathName.endsWith( File.separator ) ){
            pathName += "/";
        }
        return pathName;
    }// Tools.formatPath

    //----------------------------------------------------------------------------------------------
    /** Gets a URL from a string.  If the string begins with "jar:!",
     *  the remainder of the pathName is used to specify the file in the
     *  containing JAR file.  Otherwise, the string is used as a normal URL.
     *
     *  @param  pathName                the pathname of the file.
     *  @return URL                     - the URL from the String.
     *  @throws MalformedURLException   on any error.
     */
    public static URL getUrl( String pathName )         throws  MalformedURLException {
        return pathName.startsWith( JAR_PREFIX )
             ? new URL( classUrl, pathName.substring( JAR_PREFIX.length() - 1 ) )
             : new URL( new URL( "file:" ), (pathName.charAt( 1 ) == ':' ? "/" : "") + pathName );
    }// Tools.getUrl

    //----------------------------------------------------------------------------------------------
    /** Gets the last modified date of a file.
     *
     *  @param  urlName                 the url name of a file.
     *  @return String;                 - the last modified date of the file.
     *  @throws IOException             on any error.
     */
    public static String getLastModified( String urlName )
                                                        throws  IOException {
        URL  url  = getUrl( urlName );
        long time = url.getProtocol().equals( "jar" )
                      ? ((JarURLConnection) url.openConnection()).getJarFile()
                        .getEntry( urlName.substring( urlName.indexOf( "!/" ) + 2 ) ).getTime()
                      : url.openConnection().getLastModified();

        return time > 0 ? Log.isoDateFormat.format( new Date( time ) ) : null;
    }// Tools.getLastModified

    //----------------------------------------------------------------------------------------------
    /** Show JPopupMenu.
     *
     *  @param  jPopupMenu      the JPopupMenu to display.
     *  @param  event           the triggering MouseEvent.
     */
    public static void showMenu( JPopupMenu jPopupMenu, MouseEvent event ){
        int vPos = event.getY() - jPopupMenu.getComponentCount() * 20;

        jPopupMenu.show( event.getComponent(), event.getX(), vPos > 0 ? vPos : event.getY() );
    }// Tools.showMenu

    //----------------------------------------------------------------------------------------------
    /** Displays a resizable JOptionPane JDialog box.
     *
     *  @param  parent          the parent Component to use for the dialog box.
     *  @param  title           the String for the title of the dialog box.
     *  @param  jOptionPane     the jOptionPane to use for the dialog box.
     *  @return int             - the result from the dialog box.
     */
    public static int showDialog( Component parent, String title, JOptionPane jOptionPane ){
        JDialog jDialog = jOptionPane.createDialog( parent, title );

        jDialog.pack();
        jDialog.setResizable( true );
        jDialog.setVisible( true );
        Object result = jOptionPane.getValue();
        return result instanceof Integer        // != null
                ? (Integer) result
                : JOptionPane.CLOSED_OPTION;
    }// Tools.showDialog

    //----------------------------------------------------------------------------------------------
    /** Critical error handler.
     *
     *  @param  log             the Logger to use.
     *  @param  exception       the exception.
     */
    public static void criticalError( Logger log, Exception exception ){
        log.log( Level.SEVERE, "Critical error in configuration data", exception );
        showDialog( null, "Critical error during startup",
                    new JOptionPane( exception + Log.endOfLine + Log.endOfLine
                                               + "Error in configuration data",
                                     JOptionPane.ERROR_MESSAGE ) );
    }// Tools.criticalError

    //----------------------------------------------------------------------------------------------
    /** Log diagnostic info.
     *
     *  @param  log                             the Logger to use.
     *  @param  level                           the Logger Level to use.
     *  @param  properties                      the current Properties (may be null).
     *  @param  preferences                     the current Preferences (may be null).
     */
    @SuppressWarnings( "unchecked" )
    public static void logDiag( Logger log, Level level,
                                Properties properties, Preferences preferences ){

        // Log class versions
        try {
            Iterable <?> iterable;
            String classPath,
                   className = Tools.classUrl.getFile();

            if ( isJarFile ){
                String jarName  = className.substring( 0, className.indexOf( "!/" ) ),
                       dateTime = null;
                try {
                    dateTime = getLastModified( jarName );
                } catch ( Exception exception ){
                    log.log( Level.SEVERE, "Error getting JAR file date/time", exception );
                }
                log.log( level, "Using Java archive"
                              + (dateTime != null ? "  [created " + dateTime + "]" : "")
                              + ":  " + jarName );
                classPath = "jar:!/";
                JarURLConnection jarConnect = (JarURLConnection) classUrl.openConnection();
                iterable = new ToIterable <JarEntry>( jarConnect.getJarFile().entries() );
            } else {
                classPath = formatPath( className.substring( 0, className.lastIndexOf( '/' ) ) );
                String[] fileList  = new File( classPath ).list();
                iterable = fileList != null ? Arrays.asList( fileList ) : null;
            }
            if ( iterable != null ){
                for ( Object object : iterable ){
                    String name = classPath + object;
                    if ( name.endsWith( ".class" )
                           &&  name.lastIndexOf( '$' ) <= name.lastIndexOf( '/' ) ){
                        log.log( level, "Using class file  [created " + getLastModified( name )
                                      + "]:  "                        + name );
                    }
                }
            }
        } catch ( Exception exception ){
            log.log( Level.WARNING, "Error listing class file metadata", exception );
        }

        // Log Look and Feels
        for ( UIManager.LookAndFeelInfo lookAndFeelInfo : lookAndFeelInfos ){
            log.log( level, "Using look & feel: " + lookAndFeelInfo.getName()
                          + " = "                 + lookAndFeelInfo.getClassName() );
        }

        // Log properties
        if ( properties != null ){
            for ( String key : new ToIterable <String>( (Enumeration <String>)
                                                        properties.propertyNames() ) ){  // [unchecked]
                log.log( level, "Using property: " + key + "="
                              + properties.getProperty( key, "(no value)" ) );
            }
        }

        // Log preferences
        try {
            logPreferences( log, level, preferences );
        } catch ( Exception exception ){
            log.log( Level.WARNING, "Error listing preferences", exception );
        }
    }// Tools.logDiag

    //----------------------------------------------------------------------------------------------
    /** Recursively log preferences.
     *
     *  @param  log                             the Logger to use.
     *  @param  level                           the Logger Level to use.
     *  @param  preferences                     the current Preferences (may be null).
     *  @throws BackingStoreException           if the Backing Store is not available.
     */
    private static void logPreferences( Logger log, Level level, Preferences preferences )
                                                        throws  BackingStoreException {
        if ( preferences != null ){
            for ( String key : preferences.keys() ){
                log.log( level, "Using preference: " + preferences.absolutePath() + ":"
                              + key + "=" + preferences.get( key, "(no value)" ) );
            }
            for ( String child : preferences.childrenNames() ){
               logPreferences( log, level, preferences.node( child ) );
            }
        }
    }// Tools.logPreferences
}// Tools class
